"""
    The base functions for all objects
"""

import pdb

import data

# A private-ish class for managing the objects
# Instantiated as a singleton
class _Objects:
    list = []

    def add(self, object):
        self.list.append(object)

    def remove(self, object):
        self.list.remove(object)

    def clear(self):
        self.list = []

    def draw(self, screen):
        for obj in self.list:
            obj.draw(screen)

    def update(self, time):
        for obj in self.list:
            obj.update(time)

    def shutdown(self):
        for obj in self.list:
            obj.shutdown()
object_mgr = _Objects()

# A simple state structure
class State:
    def __init__(self, animation, time, frame):
        self.animation = animation
        self.time = time
        self.frame = frame

class Object:

    name = None                 # A plain text string describing object
    type = None                 # A constant indicating object type
    x = None                    # The x draw location
    y = None                    # The y draw location
    x_vel = None                # The current x velocity
    y_vel = None                # The current y velocity
    default_x_vel = None        # The default x velocity
    default_y_vel = None        # The default y velocity
    width = None                # width in pixels
    height = None               # height in pixels
    bound_xoffset = None        # The bounding box offset
    bound_yoffset = None        # The bounding box offset

    falling = None              # Flag for falling
    jumping = None              # Flag for jumping
    jump = None                 # The rate of upward motion for the jump
    jump_current = None         # The current jump
    jump_decay = None           # How quickly the jump decay's each cycle

    frame = None                # The current frame to draw

    state = None                # The current animation state
    animation_stack = None      # Each "stack" object should be a touple (animation, time, frame)
    animation_list = None       # The list of animations
    tileset = None              # The images to use

    def __init__(self, type):
        """Pass in a type constant (TYPE_xxxx) used in collision resolution"""
        self.type = type
        self.x = 0
        self.y = 0
        self.x_vel = 0
        self.y_vel = 0
        self.default_x_vel = 0
        self.default_y_vel = 0
        self.width = 0
        self.height = 0
        self.bound_xoffset = 0
        self.bound_yoffset = 0
        self.jump = 0
        self.jump_current = 0
        self.jump_decay = 0
        self.frame = 0
        self.state = None
        self.animation_stack = []
        self.animation_list = []
        self.tileset = []
        self.init()

    def init(self):
        """Custom init functions go here"""
        pass

    def get_x(self):
        """Return current x location"""
        return self.x

    def get_y(self):
        """Return current y location"""
        return self.y

    def get_width(self):
        """Return current x location"""
        return self.width

    def get_height(self):
        """Return current y location"""
        return self.height

    def set_name(self, name):
        """Set the human readable name of object"""
        self.name = name

    def get_name(self):
        """Return the human readable name of object"""
        return self.name

    def add_animation(self, animation_list):
        """Pass in a list of frames for an animation"""
        # validate animation list
        for frame in animation_list:
            if frame > len(self.tileset):
                print "Error, invalid animation frame: %d" % (frame)
        self.animation_list.append(animation_list)

    def load_tiles(self, filename):
        """Load a set of tiles from a file"""
        self.tileset = data.img_load(filename, tile_width=self.width, tile_height=self.height)

    def set_boundoffset(self, (x, y)):
        """Give an offset for the bounding box"""
        self.bound_xoffset = x
        self.bound_yoffest = y

    def set_frame(self, frame):
        """Set the current animation frame"""
        self.frame = frame

    def next_frame(self, time):
        """Increment to next frame
            If the animation stack has anything on it, the time will be 
        """

        new_frame = -1
        
        # Check whether we need to update state
        if self.state.frame:
            # Check to see if we are at the end of state frame
            if (self.frame == self.state.frame):
                if self.got_state():
                    self.state = self.pop_state()
                    new_frame = 0
                else:
                    new_frame = self.frame  # We are on the last state, and the last frame of that state
        elif self.state.time:
            if (time >= self.state.time):
                if self.got_state():
                    self.state = self.pop_state()
                    new_frame = 0
                else:
                    new_frame = self.frame  # We are on the last state, and the last frame of that state

        # If we haven't yet set a frame
        if 0 > new_frame:
            new_frame = self.frame + 1

            if new_frame >= len(self.animation_list[self.state.animation]):
                new_frame = 0

        self.frame = new_frame

    def next_state(self):
        """Pulls the next state from the stack, assuming there is one"""
        if self.got_state():
            self.state = self.pop_state()

    def push_state(self, animation, time, frame):
        """Push an animation onto the stack.
            If frame is set, the next state will be pop'd when frame is reached.
            If time is set, the next state will be pop'd when time is reached
            If both time and frame are set, the first reached wins
            If neither are set, animation plays indefinitely
            If there is only one animation on the stack, and frame is set, the animation will play until 'frame', 
                and 'frame' will be played indefintely.
        """

        self.animation_stack.append( State(animation, time, frame) )

    def got_state(self):
        """Returns true if there a state is availble on the stack"""
        if (0 < self.animation_stack):
            return True
        else:
            return None

    def pop_state(self):
        """Pops an animation state off the list"""
        return self.animation_stack.pop()

    def set(self, (x,y)):
        """Set the x/y location of this object. If either is None, it is not changed"""
        if None != x:
            self.x = x
        if None != y:
            self.y = y

    def set_velocity(self, xy=(0,0)):
        """Set the velocity of the object. 
            Re-sets velocity to 0 if no arguments
            This is what should be used for the actual move
        """
        self.x_vel = xy[0]
        self.y_vel = xy[1]

    def add_velocity(self, (x,y)):
        """Add x and y values to current velocity"""
        self.x_vel = self.x_vel + x
        self.y_vel = self.y_vel + y

    def move(self, (x, y), constrain_x=None, constrain_y=None):
        """Move by x/y"""
        self.x = self.x + x
        self.y = self.y + y

        if constrain_x:
            if (self.x + self.width) > constrain_x:
                self.x = constrain_x - self.width
            elif (0 > self.x):
                self.x = 0

        if constrain_y:
            if (self.y + self.width) > constrain_y:
                self.y = constrain_y - self.width
            elif (0 > self.y):
                self.y = 0

    def start_jumping(self):
        """Turn on the jumping flag"""
        if (True != self.jumping):
            self.jump_current = self.jump
            self.jumping = True

    def stop_jumping(self):
        self.jumping = None

    def start_falling(self):
        """Turn on the falling flag"""
        self.falling = True

    def stop_falling(self):
        self.falling = None

    def draw(self, screen):
        """Draw to the screen"""
        tile = self.animation_list[self.state.animation][self.frame]
        screen.blit(self.tileset[tile], (self.x, self.y))

    def jump_update(self):
        """Handles adding decay into the current jump"""
        self.jump_current = self.jump_current - self.jump_decay
        if 0 > self.jump_current:
            self.jump_current = 0

    def animation_update(self, time):
        if self.state == None:
            self.state = self.pop_state()
        self.next_frame(time)

    def update(self, time):
        print "You forgot to define update, don't forget to call animation_update()!"

    def shutdown(self):
        #print "You forgot to define shutdown!"
        pass
